@@ -2,7 +2,7 @@
 %%%
 %%% %CopyrightBegin%
 %%% 
-%%% Copyright Ericsson AB 2007-2015. All Rights Reserved.
+%%% Copyright Ericsson AB 2007-2016. All Rights Reserved.
 %%% 
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -31,11 +31,12 @@
 
 -import(hipe_tagscheme, [set_field_from_term/3, get_field_from_term/3]).
 
+-import(hipe_rtl_binary, [make_size/3]).
+
 -include("hipe_literals.hrl").
 
 %%--------------------------------------------------------------------
 
--define(MAX_BINSIZE, trunc(?MAX_HEAP_BIN_SIZE / hipe_rtl_arch:word_size()) + 2).
 -define(BYTE_SHIFT, 3). %% Turn bits into bytes or vice versa
 -define(LOW_BITS, 7). %% Three lowest bits set 
 -define(BYTE_SIZE, 8).
@@ -333,32 +334,50 @@ float_get_c_code(Dst1, Ms, Size, Flags, TrueLblName, FalseLblName) ->
 get_c_code(Func, Dst1, Ms, Size, Flags, TrueLblName, FalseLblName) ->  
   SizeReg = hipe_rtl:mk_new_reg_gcsafe(),
   FlagsReg = hipe_rtl:mk_new_reg_gcsafe(),
+  RetReg = hipe_rtl:mk_new_reg_gcsafe(),
   MatchBuf = hipe_rtl:mk_new_reg(),
   RetLabel = hipe_rtl:mk_new_label(),
+  OkLabel = hipe_rtl:mk_new_label(),
   NonVal = hipe_rtl:mk_imm(hipe_tagscheme:mk_non_value()),
   [hipe_rtl:mk_move(SizeReg, Size),
    hipe_rtl:mk_move(FlagsReg, hipe_rtl:mk_imm(Flags)),
    hipe_tagscheme:extract_matchbuffer(MatchBuf, Ms),
-   hipe_rtl_arch:call_bif([Dst1], Func, [SizeReg, FlagsReg, MatchBuf], 
+   hipe_rtl_arch:call_bif([RetReg], Func, [SizeReg, FlagsReg, MatchBuf],
 			  hipe_rtl:label_name(RetLabel), FalseLblName), 
    RetLabel,
-   hipe_rtl:mk_branch(Dst1, eq, NonVal, FalseLblName, TrueLblName, 0.01)].
+   hipe_rtl:mk_branch(RetReg, eq, NonVal, FalseLblName,
+		      hipe_rtl:label_name(OkLabel), 0.01),
+   OkLabel,
+   hipe_rtl:mk_move(Dst1, RetReg),
+   hipe_rtl:mk_goto(TrueLblName)].
 
 utf8_get_c_code(Dst, Ms, TrueLblName, FalseLblName) ->
+  RetReg = hipe_rtl:mk_new_reg_gcsafe(),
+  OkLabel = hipe_rtl:mk_new_label(),
   MatchBuf = hipe_rtl:mk_new_reg(),
   NonVal = hipe_rtl:mk_imm(hipe_tagscheme:mk_non_value()),
   [hipe_tagscheme:extract_matchbuffer(MatchBuf, Ms),
-   hipe_rtl_arch:call_bif([Dst], bs_get_utf8, [MatchBuf], [], []),
-   hipe_rtl:mk_branch(Dst, eq, NonVal, FalseLblName, TrueLblName, 0.01)].
+   hipe_rtl_arch:call_bif([RetReg], bs_get_utf8, [MatchBuf], [], []),
+   hipe_rtl:mk_branch(RetReg, eq, NonVal, FalseLblName,
+		      hipe_rtl:label_name(OkLabel), 0.01),
+   OkLabel,
+   hipe_rtl:mk_move(Dst, RetReg),
+   hipe_rtl:mk_goto(TrueLblName)].
 
 utf16_get_c_code(Flags, Dst, Ms, TrueLblName, FalseLblName) ->
+  RetReg = hipe_rtl:mk_new_reg_gcsafe(),
+  OkLabel = hipe_rtl:mk_new_label(),
   MatchBuf = hipe_rtl:mk_new_reg(),
   NonVal = hipe_rtl:mk_imm(hipe_tagscheme:mk_non_value()),
   FlagsReg = hipe_rtl:mk_new_reg_gcsafe(),
   [hipe_tagscheme:extract_matchbuffer(MatchBuf, Ms),
    hipe_rtl:mk_move(FlagsReg, hipe_rtl:mk_imm(Flags)),
-   hipe_rtl_arch:call_bif([Dst], bs_get_utf16, [MatchBuf, FlagsReg], [], []),
-   hipe_rtl:mk_branch(Dst, eq, NonVal, FalseLblName, TrueLblName, 0.01)].
+   hipe_rtl_arch:call_bif([RetReg], bs_get_utf16, [MatchBuf, FlagsReg], [], []),
+   hipe_rtl:mk_branch(RetReg, eq, NonVal, FalseLblName,
+		      hipe_rtl:label_name(OkLabel), 0.01),
+   OkLabel,
+   hipe_rtl:mk_move(Dst, RetReg),
+   hipe_rtl:mk_goto(TrueLblName)].
 
 validate_unicode_retract_c_code(Src, Ms, TrueLblName, FalseLblName) ->
   MatchBuf = hipe_rtl:mk_new_reg(),
@@ -817,10 +836,10 @@ create_lbls(0) ->
 create_lbls(X) when X > 0 ->
   [hipe_rtl:mk_new_label()|create_lbls(X-1)].
 
-make_dyn_prep(SizeReg, CCode) ->   
+make_dyn_prep(SizeReg, CCode) ->
   [CLbl, SuccessLbl] = create_lbls(2),
-  Init = [hipe_rtl:mk_branch(SizeReg, le, hipe_rtl:mk_imm(?MAX_SMALL_BITS),  
-			     hipe_rtl:label_name(SuccessLbl), 
+  Init = [hipe_rtl:mk_branch(SizeReg, leu, hipe_rtl:mk_imm(?MAX_SMALL_BITS),
+			     hipe_rtl:label_name(SuccessLbl),
 			     hipe_rtl:label_name(CLbl)),
 	  SuccessLbl],
   End = [CLbl|CCode],
@@ -865,8 +884,8 @@ get_unaligned_int_to_reg(Reg, Size, Base, Offset, LowBits, Shiftr, Type) ->
 			 hipe_rtl:mk_imm((WordSize-MinLoad)*?BYTE_SIZE))];
       {_, WordSize} ->
 	UnsignedBig = {unsigned, big},
-	[hipe_rtl:mk_branch(TotBits, le, hipe_rtl:mk_imm(MinLoad*?BYTE_SIZE), 
-			    hipe_rtl:label_name(LessLbl), 
+	[hipe_rtl:mk_branch(TotBits, leu, hipe_rtl:mk_imm(MinLoad*?BYTE_SIZE),
+			    hipe_rtl:label_name(LessLbl),
 			    hipe_rtl:label_name(MoreLbl)),
 	 LessLbl,
 	 load_bytes(LoadDst, Base, ByteOffset, Type, MinLoad),
@@ -926,7 +945,7 @@ get_big_unknown_int(Dst1, Base, Offset, NewOffset,
    hipe_rtl:mk_alu(ByteOffset, Offset, srl, hipe_rtl:mk_imm(?BYTE_SHIFT)),
    load_bytes(LoadDst, Base, ByteOffset, Type, 1),
    BackLbl,
-   hipe_rtl:mk_branch(ByteOffset, le, Limit, hipe_rtl:label_name(LoopLbl), 
+   hipe_rtl:mk_branch(ByteOffset, leu, Limit, hipe_rtl:label_name(LoopLbl),
 		      hipe_rtl:label_name(EndLbl)),
    LoopLbl,
    load_bytes(Tmp, Base, ByteOffset, {unsigned, big}, 1),
@@ -955,8 +974,8 @@ get_little_unknown_int(Dst1, Base, Offset, NewOffset,
    hipe_rtl:mk_alu(Limit, Tmp, srl, hipe_rtl:mk_imm(?BYTE_SHIFT)),
    hipe_rtl:mk_move(ShiftReg, hipe_rtl:mk_imm(0)),
    BackLbl,
-   hipe_rtl:mk_branch(ByteOffset, lt, Limit, 
-		      hipe_rtl:label_name(LoopLbl), 
+   hipe_rtl:mk_branch(ByteOffset, ltu, Limit,
+		      hipe_rtl:label_name(LoopLbl),
 		      hipe_rtl:label_name(DoneLbl)),
    LoopLbl,
    load_bytes(Tmp, Base, ByteOffset, {unsigned, big}, 1),
@@ -1072,7 +1091,7 @@ load_bytes(Dst, Base, Offset, {Signedness, Endianness}, X) when X > 1 ->
        hipe_rtl:mk_alu(Offset, Offset, add, hipe_rtl:mk_imm(1)),
        hipe_rtl:mk_alu(Dst, Dst, sll, hipe_rtl:mk_imm(8)),
        hipe_rtl:mk_alu(Dst, Dst, 'or', Tmp1),
-       hipe_rtl:mk_branch(Offset, lt, Limit, hipe_rtl:label_name(LoopLbl),
+       hipe_rtl:mk_branch(Offset, ltu, Limit, hipe_rtl:label_name(LoopLbl),
 			  hipe_rtl:label_name(EndLbl)),
        EndLbl];
     little ->
@@ -1084,7 +1103,7 @@ load_bytes(Dst, Base, Offset, {Signedness, Endianness}, X) when X > 1 ->
        hipe_rtl:mk_load(Tmp1, Base, TmpOffset, byte, Signedness),
        hipe_rtl:mk_alu(Dst, Dst, sll, hipe_rtl:mk_imm(8)),
        hipe_rtl:mk_alu(Dst, Dst, 'or', Tmp1),
-       hipe_rtl:mk_branch(Offset, lt, TmpOffset, hipe_rtl:label_name(LoopLbl),
+       hipe_rtl:mk_branch(Offset, ltu, TmpOffset, hipe_rtl:label_name(LoopLbl),
 			  hipe_rtl:label_name(EndLbl)),
        EndLbl,
        hipe_rtl:mk_move(Offset, Limit)]
@@ -1100,103 +1119,5 @@ create_gcsafe_regs(X) when X > 0 ->
 create_gcsafe_regs(0) ->
   [].
 
-first_part(Var, Register, FalseLblName) ->
-  [EndLbl] = create_lbls(1),
-  EndName = hipe_rtl:label_name(EndLbl),
-  first_part(Var, Register, FalseLblName, EndName, EndName, [EndLbl]).
-
-first_part(Var, Register, FalseLblName, TrueLblName, BigLblName, Tail) ->
-  [FixnumLbl, NotFixnumLbl, BignumLbl, SuccessLbl] = create_lbls(4),
-  [hipe_tagscheme:test_fixnum(Var, hipe_rtl:label_name(FixnumLbl),
-			      hipe_rtl:label_name(NotFixnumLbl), 0.99),
-   FixnumLbl,
-   hipe_tagscheme:fixnum_ge(Var, hipe_rtl:mk_imm(hipe_tagscheme:mk_fixnum(0)),
-			    hipe_rtl:label_name(SuccessLbl), FalseLblName,
-			    0.99),
-   SuccessLbl,
-   hipe_tagscheme:untag_fixnum(Register, Var),
-   hipe_rtl:mk_goto(TrueLblName),
-   NotFixnumLbl,
-   %% Since binaries are not allowed to be larger than 2^wordsize bits
-   %% and since bignum digits are words, we know that a bignum with an
-   %% arity larger than one can't match.
-   hipe_tagscheme:test_pos_bignum_arity(Var, 1, hipe_rtl:label_name(BignumLbl),
-					FalseLblName, 0.99),
-   BignumLbl,
-   hipe_tagscheme:unsafe_get_one_word_pos_bignum(Register, Var),
-   hipe_rtl:mk_goto(BigLblName) | Tail].
-
-make_size(1, BitsVar, FalseLblName) ->
-  [DstReg] = create_regs(1),
-  {first_part(BitsVar, DstReg, FalseLblName), DstReg};
-make_size(?BYTE_SIZE, BitsVar, FalseLblName) ->
-  [DstReg] = create_regs(1),
-  [FixnumLbl, BignumLbl] = create_lbls(2),
-  WordBits = hipe_rtl_arch:word_size() * ?BYTE_SIZE,
-  FixnumLblName = hipe_rtl:label_name(FixnumLbl),
-  Tail = [BignumLbl,
-	  hipe_rtl:mk_branch(DstReg, 'ltu',
-			     hipe_rtl:mk_imm(1 bsl (WordBits - ?BYTE_SHIFT)),
-			     FixnumLblName, FalseLblName, 0.99),
-	  FixnumLbl,
-	  hipe_rtl:mk_alu(DstReg, DstReg, sll, hipe_rtl:mk_imm(?BYTE_SHIFT))],
-  Code = first_part(BitsVar, DstReg, FalseLblName, FixnumLblName,
-		    hipe_rtl:label_name(BignumLbl), Tail),
-  {Code, DstReg};
-make_size(UnitImm, BitsVar, FalseLblName) ->
-  [DstReg] = create_regs(1),
-  UnitList = number2list(UnitImm),
-  Code = multiply_code(UnitList, BitsVar, DstReg, FalseLblName),
-  {Code, DstReg}.
-
-multiply_code(List=[Head|_Tail], Variable, Result, FalseLblName) ->
-  Test = set_high(Head),
-  Tmp1 = hipe_rtl:mk_new_reg(),
-  SuccessLbl = hipe_rtl:mk_new_label(),
-  Register = hipe_rtl:mk_new_reg(),
-  Code = [hipe_rtl:mk_move(Result, hipe_rtl:mk_imm(0))|
-	  first_part(Variable, Register, FalseLblName)]
-	++
-	 [hipe_rtl:mk_alub(Tmp1, Register, 'and', hipe_rtl:mk_imm(Test), 
-			   eq, hipe_rtl:label_name(SuccessLbl), 
-			   FalseLblName, 0.99),
-	  SuccessLbl],
-  multiply_code(List, Register, Result, FalseLblName, Tmp1, Code).
-
-multiply_code([ShiftSize|Rest], Register, Result, FalseLblName, Tmp1, OldCode) ->
-  SuccessLbl = hipe_rtl:mk_new_label(),
-  Code =
-    OldCode ++
-    [hipe_rtl:mk_alu(Tmp1, Register, sll, hipe_rtl:mk_imm(ShiftSize)),
-     hipe_rtl:mk_alub(Result, Tmp1, 'add', Result, not_overflow,
-		      hipe_rtl:label_name(SuccessLbl), FalseLblName, 0.99),
-     SuccessLbl],
-  multiply_code(Rest, Register, Result, FalseLblName, Tmp1, Code);
-multiply_code([], _Register, _Result, _FalseLblName, _Tmp1, Code) ->
-  Code.
-
-number2list(X) when is_integer(X), X >= 0 ->
-  number2list(X, []).
-
-number2list(1, Acc) ->
-  lists:reverse([0|Acc]);
-number2list(0, Acc) ->
-  lists:reverse(Acc);
-number2list(X, Acc) ->
-  F = floorlog2(X),
-  number2list(X-(1 bsl F), [F|Acc]).
-
-floorlog2(X) ->
-  round(math:log(X)/math:log(2)-0.5). 
-
-set_high(X) ->
-  WordBits = hipe_rtl_arch:word_size() * ?BYTE_SIZE,
-  set_high(min(X, WordBits), WordBits, 0).
-
-set_high(0, _, Y) ->
-  Y;
-set_high(X, WordBits, Y) ->
-  set_high(X-1, WordBits, Y+(1 bsl (WordBits-X))).
-
 is_illegal_const(Const) ->
   Const >=  1 bsl (hipe_rtl_arch:word_size() * ?BYTE_SIZE) orelse Const < 0.
